// Source file for the mesh to grid conversion program



// Include files 

#include "R3Shapes/R3Shapes.h"
#include "R3Harmonics/R3Harmonics.h"



// Program variables

static char *mesh_name = NULL;
static char *points_name = NULL;
static char *shd_name = NULL;
static R3Point center_point(0, 0, 0);
static RNBoolean center_initialized = FALSE;
static RNScalar world_radius_scale = 1.0;
static int max_grid_resolution = 512;
static int descriptor_type = 0; // 0=shells, 1=shd, 2=sfd, 3=sgd
static RNLength gedt_sigma = sqrt(8);
static int normalization_type = 0;
static int nshells = 32;
static int bandwidth = 16;
static RNBoolean write_header = FALSE;
static int print_verbose = 0;
static int print_debug = 0;



// Type definitions

struct Point {
  R3Point position;
  R3Vector normal;
};



static R3Point
MeshCentroid(R3Mesh *mesh)
{
  // Compute mesh center of mass
  RNScalar mesh_area = 0.0;
  R3Point mesh_centroid = R3zero_point;
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);
    RNArea area = mesh->FaceArea(face);
    mesh_centroid += area * mesh->FaceCentroid(face);
    mesh_area += area;
  }
  mesh_centroid /= mesh_area;

  // Return centroid
  return mesh_centroid;
}



static RNLength
MeshRadius(R3Mesh *mesh)
{
  // Compute mesh center of mass
  RNScalar mesh_area = 0.0;
  R3Point mesh_centroid = R3zero_point;
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);
    RNArea area = mesh->FaceArea(face);
    mesh_centroid += area * mesh->FaceCentroid(face);
    mesh_area += area;
  }
  mesh_centroid /= mesh_area;

  // Compute average distance to center of mass
  RNScalar avg_distance = 0.0;
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);
    RNArea area = mesh->FaceArea(face);
    avg_distance += area * R3Distance(mesh->FaceCentroid(face), mesh_centroid);
  }
  avg_distance /= mesh_area;

  // Compute world radius
  RNScalar mesh_radius = 2.5 * avg_distance;

  // Return world radius
  return mesh_radius;
}



static R3Mesh *
ReadMesh(char *mesh_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate mesh
  R3Mesh *mesh = new R3Mesh();
  assert(mesh);

  // Read mesh from file
  if (!mesh->ReadFile(mesh_name)) {
    delete mesh;
    return NULL;
  }

  // Print statistics
  if (print_verbose) {
    printf("Read mesh ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Faces = %d\n", mesh->NFaces());
    printf("  # Edges = %d\n", mesh->NEdges());
    printf("  # Vertices = %d\n", mesh->NVertices());
    fflush(stdout);
  }

  // Return success
  return mesh;
}



static RNArray<Point *> *
ReadPoints(const char *filename)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate array of points
  RNArray<Point *> *points = new RNArray<Point *>();
  if (!points) {
    fprintf(stderr, "Unable to allocate array of points\n");
    return NULL;
  }

  // Open points file
  FILE *fp = fopen(filename, "rb");
  if (!fp) {
    fprintf(stderr, "Unable to open points file: %s\n", filename);
    return NULL;
  }

  // Read points
  float coordinates[6];
  while (fread(coordinates, sizeof(float), 6, fp) == (unsigned int) 6) {
    Point *point = new Point();
    if (!point) { 
      fprintf(stderr, "Unable to create point %d for %s\n", points->NEntries(), filename); 
      return NULL; 
    }
    point->position.Reset(coordinates[0], coordinates[1], coordinates[2]);
    point->normal.Reset(coordinates[3], coordinates[4], coordinates[5]);
    points->Insert(point);
  }

  // Close points file
  fclose(fp);

  // Print statistics
  if (print_verbose) {
    printf("Read point file: %s ...\n", filename);
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", points->NEntries());
    fflush(stdout);
  }

  // Return points
  return points;
}



static R3Grid *
CreateGrid(R3Mesh *mesh, const RNArray<Point *> *points,
  RNLength world_radius, RNLength grid_spacing, int max_resolution)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Compute bounding box of points
  R3Box points_bbox(R3null_box);
  for (int i = 0; i < points->NEntries(); i++) {
    Point *point = points->Kth(i);
    points_bbox.Union(point->position);
  }

  // Compute bounding box of all descriptors (plus four voxels on all sides)
  R3Vector extension = (world_radius + 4 * grid_spacing) * R3ones_vector;
  R3Box descriptors_bbox(points_bbox.Min() - extension, points_bbox.Max() + extension);

  // Compute grid resolution
  int xres = (int) (descriptors_bbox.XLength() / grid_spacing);
  int yres = (int) (descriptors_bbox.YLength() / grid_spacing);
  int zres = (int) (descriptors_bbox.ZLength() / grid_spacing);
  if (xres > max_resolution) xres = max_resolution;
  if (yres > max_resolution) yres = max_resolution;
  if (zres > max_resolution) zres = max_resolution;

  // Allocate grid
  R3Grid *grid = new R3Grid(xres, yres, zres);
  if (!grid) {
    fprintf(stderr, "Unable to allocate grid\n");
    exit(-1);
  }

  // Compute grid transformation
  grid->SetWorldToGridTransformation(descriptors_bbox);

  // Rasterize each triangle into grid
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);
    const R3Point& p0 = mesh->VertexPosition(mesh->VertexOnFace(face, 0));
    const R3Point& p1 = mesh->VertexPosition(mesh->VertexOnFace(face, 1));
    const R3Point& p2 = mesh->VertexPosition(mesh->VertexOnFace(face, 2));
    grid->RasterizeWorldTriangle(p0, p1, p2, 1.0);
  }

  // Compute the squared distance transform
  grid->SquaredDistanceTransform();

  // Print statistics
  if (print_verbose) {
    printf("Created grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Points BBox = ( %g %g %g ) ( %g %g %g )\n", 
      points_bbox[0][0], points_bbox[0][1], points_bbox[0][2], 
      points_bbox[1][0], points_bbox[1][1], points_bbox[1][2]);
    printf("  Descriptors BBox = ( %g %g %g ) ( %g %g %g )\n", 
      descriptors_bbox[0][0], descriptors_bbox[0][1], descriptors_bbox[0][2], 
      descriptors_bbox[1][0], descriptors_bbox[1][1], descriptors_bbox[1][2]);
    printf("  BBox Extension = %g\n", extension.X());
    printf("  World radius = %g\n", world_radius);
    printf("  World spacing = %g ( %g )\n", grid_spacing, grid->GridToWorldScaleFactor());
    printf("  Grid resolution = %d %d% d\n", grid->XResolution(), grid->YResolution(), grid->ZResolution());
    printf("  Cardinality = %d\n", grid->Cardinality());
    RNInterval range = grid->Range();
    printf("  Minimum = %g\n", range.Min());
    printf("  Maximum = %g\n", range.Max());
    printf("  L1Norm = %g\n", grid->L1Norm());
    printf("  L2Norm = %g\n", grid->L2Norm());
    fflush(stdout);
  }

  // Return grid
  return grid;
}



static R3SphericalGrid *
CreateSphericalGrid(R3Grid *grid1, const R3Point& world_center, 
  int nshells, int bandwidth, RNScalar world_radius, RNLength gedt_sigma)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate spherical grid
  R3SphericalGrid *grid = new R3SphericalGrid(bandwidth * 2, bandwidth * 2, nshells);
  if (!grid) {
    fprintf(stderr, "Unable to allocate spherical grid\n");
    exit(-1);
  }

  // Set world-to-grid transformation
  grid->SetWorldToGridTransformation(world_center, R3negz_vector, R3posy_vector, world_radius);

  // Fill samples on spherical grid with Gaussian of DT
  const RNScalar sqrt_two_pi = sqrt(RN_TWO_PI);
  RNScalar a = sqrt_two_pi * gedt_sigma;
  RNScalar fac = 1.0 / (a * a * a);
  RNScalar denom = -2.0 * gedt_sigma * gedt_sigma;
  for (int k = 0; k < grid->RadiusResolution(); k++) {
    for (int j = 0; j < grid->ThetaResolution(); j++) {
      for (int i = 0; i < grid->PhiResolution(); i++) {
        R3Point world_position = grid->WorldPosition(i, j, k);
        RNScalar value1 = grid1->WorldValue(world_position);
        RNScalar value = fac * exp( value1 / denom );
        grid->SetGridValue(i, j, k, value);
      }
    }
  }

  // Print statistics
  if (print_debug) {
    printf("Created spherical grid ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Voxels = %d %d% d\n", grid->PhiResolution(), grid->ThetaResolution(), grid->RadiusResolution());
    printf("  L2Norm = %g\n", grid->L2Norm());
    printf("  World center = %g %g %g\n", world_center[0], world_center[1], world_center[2]);
    printf("  World radius = %g\n", world_radius);
    printf("  Sigma = %g\n", gedt_sigma);
    fflush(stdout);
  }

  // Return spherical grid
  return grid;
}



static R3SphericalHarmonics *
CreateSphericalHarmonics(R3SphericalGrid *grid)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Allocate spherical harmonics
  R3SphericalHarmonics *harmonics = new R3SphericalHarmonics(grid->ThetaResolution()/2, grid->RadiusResolution());
  if (!harmonics) {
    fprintf(stderr, "Unable to allocate spherical harmonics\n");
    exit(-1);
  }

  // Compute forward harmonic transform
  ForwardSphericalHarmonicTransform(*grid, *harmonics);

  // Print statistics
  if (print_debug) {
    printf("Created spherical harmonics ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Coefficients = %d\n", harmonics->NCoefficients());
    printf("  # Shells = %d\n", harmonics->RadiusResolution());
    printf("  Bandwidth = %d\n", harmonics->Bandwidth());
    printf("  L2Norm = %g\n", harmonics->L2Norm());
    fflush(stdout);
  }

  // Return spherical harmonics
  return harmonics;
}



static int
CreateDescriptor(R3SphericalHarmonics *harmonics, int descriptor_type, float *values)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();
  float sum = 0.0;

  // Initialize counts
  int nvalues = 0;
  int count = 0;

  // Check descriptor type
  if (descriptor_type == 0) {
    // Compute number of values
    nvalues = harmonics->RadiusResolution();

    // Compute L2Norm for each radius 
    for (int S = 0; S < harmonics->RadiusResolution(); S++) {
      float radius = S + 0.5;
      float scale_factor = 4.0 * RN_PI * radius * radius;
      float scale_factor_squared = scale_factor * scale_factor;
      float value_squared = scale_factor_squared * harmonics->L2NormSquared(S);
      float value = sqrt(value_squared);
      assert(count <= nvalues);
      values[count++] = value;
      sum += value_squared;
    }
  }
  else if (descriptor_type == 1) {
    // Compute number of values
    nvalues = harmonics->RadiusResolution() * harmonics->Bandwidth();

    // Compute L2Norm for each radius and frequency
    for (int S = 0; S < harmonics->RadiusResolution(); S++) {
      float radius = S + 0.5;
      float scale_factor = 4.0 * RN_PI * radius * radius;
      float scale_factor_squared = scale_factor * scale_factor;
      for (int L = 0; L < harmonics->Bandwidth(); L++) {
        float value_squared = scale_factor_squared * harmonics->L2NormSquared(S, L);
        float value = sqrt(value_squared);
        assert(count <= nvalues);
        values[count++] = value;
        sum += value_squared;
      }
    }
  }
  else if (descriptor_type == 2) {
    // Compute number of values
    nvalues = harmonics->NCoefficients() / 2;

    // Compute L2Norm for each coefficient
    for (int S = 0; S < harmonics->RadiusResolution(); S++) {
      float radius = S + 0.5;
      float scale_factor = 4.0 * RN_PI * radius * radius;
      float scale_factor_squared = scale_factor * scale_factor;
      for (int L = 0; L < harmonics->Bandwidth(); L++) {
        for (int M = -L; M <= L; M++) {
          float value_squared = scale_factor_squared * harmonics->L2NormSquared(S, L, M);
          float value = sqrt(value_squared);
          assert(count <= nvalues);
          values[count++] = value;
          sum += value_squared;
        }
      }
    }
  }
  else if (descriptor_type == 3) {
    // Compute number of values
    nvalues = harmonics->NCoefficients();

    // Compute coefficients for each radius and frequency
    for (int S = 0; S < harmonics->RadiusResolution(); S++) {
      float radius = S + 0.5;
      float scale_factor = 4.0 * RN_PI * radius * radius;
      for (int L = 0; L < harmonics->Bandwidth(); L++) {
        for (int M = -L; M <= L; M++) {
          for (int k = 0; k < 2; k++) {
            float value = scale_factor * harmonics->Shell(S).Coefficient(L, M, k);
            float value_squared = value * value;
            assert(count <= nvalues);
            values[count++] = value;
            sum += value_squared;
          }
        }
      }
    }
  }
  else {
    RNAbort("Invalid descriptor type");
  }

  // Compute l2norm
  assert(nvalues > 0);
  assert(count == nvalues);
  RNScalar l2norm = sqrt(sum);

  // Normalize descriptor
  if ((normalization_type % 2) == 1) {
    if (l2norm > 0) {
      for (int i = 0; i < nvalues; i++) {
        values[i] /= l2norm;
      }
    }
  }

  // Print statistics
  if (print_debug) {
    printf("Created spherical harmonic descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Values = %d\n", nvalues);
    printf("  # Shells = %d\n", harmonics->RadiusResolution());
    printf("  Bandwidth = %d\n", harmonics->Bandwidth());
    printf("  L2Norm = %g\n", l2norm);
    fflush(stdout);
  }

  // Return number of values
  return nvalues;
}



static int
NormalizeDescriptor(int nvalues, float *values, int normalization_type)
{
  // Check if normalizing
  if (normalization_type < 2) return 1;

  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Compute l2norm
  float l2norm = 0;
  for (int i = 0; i < nvalues; i++)
    l2norm += values[i] * values[i];
  l2norm = sqrt(l2norm);

  // Normalize descriptor
  if (l2norm > 0) {
    for (int i = 0; i < nvalues; i++)
      values[i] /= l2norm;
  }

  // Print statistics
  if (print_debug) {
    printf("Normalized spherical harmonic descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  Normalization type = %d\n", normalization_type);
    printf("  L2Norm = %g\n", l2norm);
    fflush(stdout);
  }

  // Return success
  return 1;
}



static int
WriteDescriptor(int nvalues, float *values, FILE *fp)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Write number of values
  if (fwrite(&nvalues, sizeof(int), 1, fp) != 1) {
    fprintf(stderr, "Unable to write number of values to SHD file %s\n", shd_name);
    return 0;
  }

  // Write values
  if (fwrite(values, sizeof(float), nvalues, fp) != (unsigned long) nvalues) {
    fprintf(stderr, "Unable to write values to SHD file %s\n", shd_name);
    return 0;
  }

  // Print statistics
  if (print_debug) {
    printf("Wrote spherical harmonic descriptor ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Bytes = %d\n", sizeof(int) + nvalues * sizeof(float));
    fflush(stdout);
  }

  // Return number of values written
  return nvalues;
}



static int
WriteDescriptors(R3Grid *grid, const RNArray<Point *> *points, RNBoolean write_header,
  int nshells, int bandwidth, RNLength world_radius, RNLength gedt_sigma, 
  const char *shd_name)
{
  // Start statistics
  RNTime start_time;
  start_time.Read();

  // Open output file
  FILE *shd_file = fopen(shd_name, "wb");
  if (!shd_file) {
    fprintf(stderr, "Unable to open SHD file %s\n", shd_name);
    return 0;
  }

  // Write header
  if (write_header) {
    // Write magic key to identify header
    const int magic = -87654321;
    if (fwrite(&magic, sizeof(int), 1, shd_file) != 1) {
      fprintf(stderr, "Unable to write %s\n", shd_file);
      return 0;
    }

    // Write parameters
    int dummy[8] = { 0 };
    int npoints = points->NEntries();
    fwrite(&npoints, sizeof(int), 1, shd_file);
    fwrite(&descriptor_type, sizeof(int), 1, shd_file);
    fwrite(&world_radius, sizeof(RNLength), 1, shd_file);
    fwrite(&world_radius_scale, sizeof(RNScalar), 1, shd_file);
    fwrite(&nshells, sizeof(int), 1, shd_file);
    fwrite(&bandwidth, sizeof(int), 1, shd_file);
    fwrite(&gedt_sigma, sizeof(RNLength), 1, shd_file);
    fwrite(&dummy, sizeof(int), 8, shd_file);

    // Write points
    float coordinates[6];
    for (int i = 0; i < points->NEntries(); i++) {
      Point *point = points->Kth(i);
      coordinates[0] = point->position.X();
      coordinates[1] = point->position.Y();
      coordinates[2] = point->position.Z();
      coordinates[3] = point->normal.X();
      coordinates[4] = point->normal.Y();
      coordinates[5] = point->normal.Z();
      fwrite(coordinates, sizeof(float), 6, shd_file);
    }
  }

  // Compute/concatenate/write SHDs for all points
  for (int i = 0; i < points->NEntries(); i++) {
    Point *point = points->Kth(i);
    const R3Point& position = point->position;
    // const R3Point& normal = point->normal;

    // Create spherical grid
    R3SphericalGrid *sgrid = CreateSphericalGrid(grid, position, nshells, bandwidth, world_radius, gedt_sigma);
    if (!sgrid) exit(-1);

    // Create spherical harmonics
    R3SphericalHarmonics *harmonics = CreateSphericalHarmonics(sgrid);
    if (!harmonics) exit(-1);

    // Create spherical harmonic descriptor
    const int max_descriptor_size = 16 * 1024 * 1024;
    static float descriptor[max_descriptor_size];
    int descriptor_size = CreateDescriptor(harmonics, descriptor_type, descriptor);
    assert(descriptor_size < max_descriptor_size);
    if (descriptor_size == 0) exit(-1);

    // Normalize descriptor
    int status1 = NormalizeDescriptor(descriptor_size, descriptor, normalization_type);
    if (!status1) exit(-1);

    // Write descriptor
    int status2 = WriteDescriptor(descriptor_size, descriptor, shd_file);
    if (!status2) exit(-1);
  }

  // Close file
  fclose(shd_file);

  // Print statistics
  if (print_verbose) {
    printf("Created spherical harmonic descriptors ...\n");
    printf("  Time = %.2f seconds\n", start_time.Elapsed());
    printf("  # Points = %d\n", points->NEntries());
    printf("  # Shells = %d\n", nshells);
    printf("  Bandwidth = %d\n", bandwidth);
    printf("  World radius = %g\n", world_radius);
    printf("  World radius scale = %g\n", world_radius_scale);
    printf("  Sigma = %g\n", gedt_sigma);
    printf("  Normalization = %d\n", normalization_type);
    printf("  Descriptor type = %d\n", descriptor_type);
    printf("  Header = %d\n", (write_header) ? 1 : 0);
    fflush(stdout);
  }

  // Return number of points
  return points->NEntries();
}



static int 
ParseArgs(int argc, char **argv)
{
  // Check number of arguments
  if (argc < 3) {
    printf("Usage: off2shd meshfile shdfile [-points pointsfile] [-v]\n");
    exit(0);
  }

  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) print_verbose = 1; 
      else if (!strcmp(*argv, "-debug")) print_debug = 1; 
      else if (!strcmp(*argv, "-header")) { write_header = TRUE; }
      else if (!strcmp(*argv, "-points")) { argc--; argv++; points_name = *argv; }
      else if (!strcmp(*argv, "-max_resolution")) { argc--; argv++; max_grid_resolution = atoi(*argv); }
      else if (!strcmp(*argv, "-nshells")) { argc--; argv++; nshells = atoi(*argv); }
      else if (!strcmp(*argv, "-bandwidth")) { argc--; argv++; bandwidth = atoi(*argv); }
      else if (!strcmp(*argv, "-scale")) { argc--; argv++; world_radius_scale = atof(*argv); }
      else if (!strcmp(*argv, "-sigma")) { argc--; argv++; gedt_sigma = atof(*argv); }
      else if (!strcmp(*argv, "-normalize")) { argc--; argv++; normalization_type = atoi(*argv); }
      else if (!strcmp(*argv, "-sig")) { argc--; argv++; descriptor_type = atoi(*argv); }
      else if (!strcmp(*argv, "-center")) { 
        argc--; argv++; center_point[0] = atof(*argv); 
        argc--; argv++; center_point[1] = atof(*argv); 
        argc--; argv++; center_point[2] = atof(*argv); 
        center_initialized = TRUE;
      }
      else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
    }
    else {
      if (!mesh_name) mesh_name = *argv;
      else if (!shd_name) shd_name = *argv;
      else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
    }
    argv++; argc--;
  }

  // Check mesh filename
  if (!mesh_name) {
    fprintf(stderr, "You did not specify a mesh file.\n");
    return 0;
  }

  // Check shd filename
  if (!shd_name) {
    fprintf(stderr, "You did not specify a signature file.\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



int 
main(int argc, char **argv)
{
  // Parse program arguments
  if (!ParseArgs(argc, argv)) exit(-1);

  // Read mesh file
  R3Mesh *mesh = ReadMesh(mesh_name);
  if (!mesh) exit(-1);

  // Read/create points
  RNArray<Point *> *points = NULL;
  if (points_name) {
    // Read points file
    points = ReadPoints(points_name);
    if (!points) exit(-1);
  }
  else {
    // Create one point (at centroid by default)
    points = new RNArray<Point *>();
    Point *point = new Point();
    if (center_initialized) point->position = center_point;
    else point->position = MeshCentroid(mesh);
    point->normal = R3zero_vector;
    points->Insert(point);
  }

  // Compute grid parameters
  RNLength world_radius = world_radius_scale * MeshRadius(mesh);
  RNLength grid_spacing = world_radius / nshells;

  // Create grid from mesh
  R3Grid *grid = CreateGrid(mesh, points, world_radius, grid_spacing, max_grid_resolution);
  if (!grid) exit(-1);

  // Write descriptors
  int status = WriteDescriptors(grid, points, write_header, nshells, bandwidth, world_radius, gedt_sigma, shd_name);
  if (!status) exit(-1);

  // Return success
  return 0;
}

