// Include files

#include "R3Harmonics/R3Harmonics.h"



void
R3EulerRotationIndices(int index, int size, int& i, int& j, int& k)
{
  // Just checking
  assert((index >= 0) && (index < size*size*size));

  // Compute euler grid indices
  int size_squared = size * size;
  j = index / size_squared;
  i = (index - j * size_squared) / size;
  k = index % size;

  // Just checking
  assert((i >= 0) && (i < size));
  assert((j >= 0) && (j < size));
  assert((k >= 0) && (k < size));
}



void 
R3EulerRotationAngles(int i, int j, int k, int size, R3Vector& euler_angles)
{
  // Just checking
  assert((i >= 0) && (i < size));
  assert((j >= 0) && (j < size));
  assert((k >= 0) && (k < size));

  // Compute rotation angles
  RNAngle gamma=(2.0*RN_PI*i)/size;
  RNAngle beta=RN_PI*(2.0*j+1)/(2.0*size);
  RNAngle alpha=(2.0*RN_PI*k)/size;
  euler_angles.Reset(alpha, beta, gamma);
}



void
R3EulerRotationAngles(int index, int size, R3Vector& euler_angles)
{
  // Just checking
  assert((index >= 0) && (index < size*size*size));

  // Fill euler angles for index in euler grid of size
  int i, j, k;
  R3EulerRotationIndices(index, size, i, j, k);
  R3EulerRotationAngles(i, j, k, size, euler_angles);
}



void
R3EulerRotationMatrix(const R3Vector& euler_angles, R4Matrix& rotation_matrix)
{
  // Fill rotation matrix
  rotation_matrix = R4identity_matrix;
  rotation_matrix.YRotate(euler_angles[0]);
  rotation_matrix.ZRotate(euler_angles[1]);
  rotation_matrix.YRotate(euler_angles[2]);
}



void
R3EulerRotationMatrix(int i, int j, int k, int size, R4Matrix& rotation_matrix)
{
  // Just checking
  assert((i >= 0) && (i < size));
  assert((j >= 0) && (j < size));
  assert((k >= 0) && (k < size));

  // Get rotation angles
  R3Vector euler_angles;
  R3EulerRotationAngles(i, j, k, size, euler_angles);
  R3EulerRotationMatrix(euler_angles, rotation_matrix);
}



void
R3EulerRotationMatrix(int index, int size, R4Matrix& rotation_matrix)
{
  // Just checking
  assert((index >= 0) && (index < size*size*size));

  // Get rotation angles
  R3Vector euler_angles;
  R3EulerRotationAngles(index, size, euler_angles);
  R3EulerRotationMatrix(euler_angles, rotation_matrix);
}



RNInterval
R2CorrelationForAllRotations(const R2SphericalHarmonics& harmonics1, const R2SphericalHarmonics& harmonics2, 
  double *euler_grid, int *min_correlation_index, int *max_correlation_index)
{
  // Just checking
  int bw = harmonics1.Bandwidth();
  assert(harmonics1.Bandwidth() == harmonics2.Bandwidth());

  // Create a WignerD structure
  R3WignerD wigner(bw);

  // Add spherical harmonics correlations to the WignerD structure
  wigner.AddSphericalHarmonics(harmonics1, harmonics2);

  // Create an euler grid (if there is none)
  int size = 2 * bw;
  double *grid = euler_grid;
  if (!euler_grid) { 
    grid = new RNScalar [size * size * size]; 
    assert(grid); 
  }

  // Take the inverse WignerD transform
  wigner.InverseWignerDTransform(grid);

  // Search for the rotations with minimum and maximum correlation
  int min_index = 0;
  int max_index = 0;
  RNScalar min_correlation = FLT_MAX;
  RNScalar max_correlation = -FLT_MAX;
  for (int i = 0; i < size * size * size; i++) {
    if (grid[i] < min_correlation) {
      min_correlation = grid[i];
      min_index = i;
    }
    if (grid[i] > max_correlation) {
      max_correlation = grid[i];
      max_index = i;
    }
  }

  // Fill output variables
  if (min_correlation_index) *min_correlation_index = min_index;
  if (max_correlation_index) *max_correlation_index = max_index;

  // Delete temporary variables
  if (!euler_grid) delete [] grid;

  // Return interval of correlation values
  return RNInterval(min_correlation, max_correlation);
}



RNInterval
R2CorrelationForAllRotations(const R2SphericalGrid& grid1, const R2SphericalGrid& grid2,
  double *euler_grid, int *min_correlation_index, int *max_correlation_index)
{
  // Allocate spherical harmonic coefficients for both grids
  R2SphericalHarmonics harmonics1(grid1.ThetaResolution()/2);
  R2SphericalHarmonics harmonics2(grid2.ThetaResolution()/2);

  // Compute forward harmonic transform for both grids
  grid1.ForwardSphericalHarmonicTransform(harmonics1);
  grid2.ForwardSphericalHarmonicTransform(harmonics2);

  // Compute correlation for all rotations
  return R2CorrelationForAllRotations(harmonics1, harmonics2, 
    euler_grid, min_correlation_index, max_correlation_index);
}



RNInterval
R3CorrelationForAllRotations(const R3SphericalHarmonics& harmonics1, const R3SphericalHarmonics& harmonics2, 
  double *euler_grid, int *min_correlation_index, int *max_correlation_index)
{
  // Just checking
  int bw = harmonics1.Bandwidth();
  int nshells = harmonics1.RadiusResolution();
  assert(harmonics1.Bandwidth() == harmonics2.Bandwidth());
  assert(harmonics1.RadiusResolution() == harmonics2.RadiusResolution());

  // Create a WignerD structure
  R3WignerD wigner(bw);

  // Add spherical harmonics correlations to the WignerD structure
  for (int i = 0; i < nshells; i++) {
    wigner.AddSphericalHarmonics(harmonics1.Shell(i), harmonics2.Shell(i));
  }

  // Create an euler grid (if there is none)
  int size = 2 * bw;
  double *grid = euler_grid;
  if (!euler_grid) { 
    grid = new double [size * size * size]; 
    assert(grid); 
  }

  // Take the inverse WignerD transform
  wigner.InverseWignerDTransform(grid);

  // Search for the rotations with minimum and maximum correlation
  int min_index = 0;
  int max_index = 0;
  RNScalar min_correlation = FLT_MAX;
  RNScalar max_correlation = -FLT_MAX;
  for (int i = 0; i < size * size * size; i++) {
    if (grid[i] < min_correlation) {
      min_correlation = grid[i];
      min_index = i;
    }
    if (grid[i] > max_correlation) {
      max_correlation = grid[i];
      max_index = i;
    }
  }

  // Fill output variables
  if (min_correlation_index) *min_correlation_index = min_index;
  if (max_correlation_index) *max_correlation_index = max_index;

  // Delete temporary variables
  if (!euler_grid) delete [] grid;

  // Return interval of correlation values
  return RNInterval(min_correlation, max_correlation);
}



RNInterval
R3CorrelationForAllRotations(const R3SphericalGrid& grid1, const R3SphericalGrid& grid2,
  double *euler_grid, int *min_correlation_index, int *max_correlation_index)
{
  // Allocate spherical harmonic coefficients for both grids
  R3SphericalHarmonics harmonics1(grid1.ThetaResolution()/2, grid1.RadiusResolution());
  R3SphericalHarmonics harmonics2(grid2.ThetaResolution()/2, grid2.RadiusResolution());

  // Compute forward harmonic transform for both grids
  grid1.ForwardSphericalHarmonicTransform(harmonics1);
  grid2.ForwardSphericalHarmonicTransform(harmonics2);

  // Compute correlation for all rotations
  return R3CorrelationForAllRotations(harmonics1, harmonics2, 
    euler_grid, min_correlation_index, max_correlation_index);
}



RNInterval
R3CorrelationForAllRotations(const R3Grid& grid1, const R3Grid& grid2,
  double *euler_grid, int *min_correlation_index, int *max_correlation_index)
{
  // Create spherical grids
  R3SphericalGrid sgrid1(grid1.XResolution(), grid1.XResolution(), grid1.XResolution()/2);
  R3SphericalGrid sgrid2(grid2.XResolution(), grid2.XResolution(), grid2.XResolution()/2);

  // Set world-to-grid transformations
  sgrid1.SetWorldToGridTransformation(grid1.WorldBox());
  sgrid2.SetWorldToGridTransformation(grid2.WorldBox());

  // Fill the spherical grids
  sgrid1.Add(grid1);
  sgrid2.Add(grid2);

  // Compute correlation for all rotations
  return R3CorrelationForAllRotations(sgrid1, sgrid2, 
    euler_grid, min_correlation_index, max_correlation_index);
}



