// Source file for GAPS WignerD transform class



////////////////////////////////////////////////////////////////////////
// This class stores WignerD coefficients.
// The coefficients are arranged in a way consistent with the
// soft10 library.
////////////////////////////////////////////////////////////////////////



// Include files

#include "R3Harmonics/R3Harmonics.h"



R3WignerD::
R3WignerD(int bw)
  : bandwidth(bw),
    ncoefficients( 2 * (4 * bw * bw * bw - bw ) / 3 ),
    coefficients(NULL)
{
  // Allocate array for coefficients
  if (bandwidth >= 0) {
    coefficients = new double [ ncoefficients ];
    assert(coefficients);
  }
}



R3WignerD::
R3WignerD(const R3WignerD& w)
  : bandwidth(w.bandwidth),
    ncoefficients(w.ncoefficients),
    coefficients(NULL)
{
  // Allocate/copy array for coefficients
  if (bandwidth >= 0) {
    // Allocate memory
    coefficients = new double [ ncoefficients ];
    assert(coefficients);

    // Copy coefficients
    for (int i = 0; i < ncoefficients; i++) 
      coefficients[i] = w.coefficients[i];
  }
}



R3WignerD::
~R3WignerD(void)
{
  // Delete coefficients
  if (coefficients) delete [] coefficients;
}



void R3WignerD::
Clear(void)
{
  // Set all coefficients to zero
  for (int i = 0; i < ncoefficients; i++) 
    coefficients[i] = 0;
}



void R3WignerD::
SetRealCoefficient(int L, int M1, int M2, RNScalar value)
{
  // Set coefficient
  int index = Index(L, M1, M2);
  assert((index >= 0) < (index < ncoefficients/2));
  coefficients[index] = value;
}



void R3WignerD::
SetImaginaryCoefficient(int L, int M1, int M2, RNScalar value)
{
  // Set coefficient
  int index = Index(L, M1, M2);
  assert((index >= 0) < (index < ncoefficients/2));
  coefficients[ncoefficients/2 + index] = value;
}



int R3WignerD::
ReadFile(const char *filename)
{
  // Open file
  FILE *fp = fopen(filename, "rb");
  if (!fp) {
    RNFail("Unable to open file %s", filename);
    return 0;
  }

  // Read 
  int status = Read(fp);

  // Close file
  fclose(fp);

  // Return status
  return status;
}



int R3WignerD::
WriteFile(const char *filename) const
{
  // Open file
  FILE *fp = fopen(filename, "wb");
  if (!fp) {
    RNFail("Unable to open file %s", filename);
    return 0;
  }

  // Read 
  int status = Write(fp);

  // Close file
  fclose(fp);

  // Return status
  return status;
}



int R3WignerD::
Read(FILE *fp)
{
  // Check file
  if (!fp) fp = stdin;

  // Read bandwidth from file
  if (fread(&bandwidth, sizeof(int), 1, fp) != 1) {
    RNFail("Unable to read bandwidth from WignerD file");
    return 0;
  }

  // Compute number of coefficients
  ncoefficients = 2 * (4 * bandwidth * bandwidth * bandwidth - bandwidth ) / 3;

  // Allocate coefficients
  coefficients = new RNScalar [ ncoefficients ];
  assert(coefficients);

  // Read coefficients
  if (fread(coefficients, sizeof(RNScalar), ncoefficients, fp) != (unsigned int) ncoefficients) {
    RNFail("Unable to read %d coefficients from WignerD file", ncoefficients);
    return 0;
  }

  // Return number of coefficients read
  return ncoefficients;
}



int R3WignerD::
Write(FILE *fp) const
{
  // Check file
  if (!fp) fp = stdout;

  // Write bandwidth to file
  if (fwrite(&bandwidth, sizeof(int), 1, fp) != 1) {
    RNFail("Unable to write bandwidth to WignerD file");
    return 0;
  }

  // Write coefficients
  if (fwrite(coefficients, sizeof(RNScalar), ncoefficients, fp) != (unsigned int) ncoefficients) {
    RNFail("Unable to write coefficients to WignerD file");
    return 0;
  }

  // Return number of coefficients written
  return ncoefficients;
}



int R3WignerD::
Print(FILE *fp) const
{
  // Check file
  if (!fp) fp = stdout;

  // Print coefficients
  for (int L = 0; L < bandwidth; L++) {
    for (int M1 = -L ; M1 <= L; M1++) {
      for (int M2 = -L ; M2 <= L; M2++) {
        fprintf(fp, "( %g %g )", RealCoefficient(L, M1, M2), ImaginaryCoefficient(L, M1, M2));
      }
      fprintf(fp, "\n");
    }
    fprintf(fp, "\n");
  }

  // Return number of coefficients written
  return ncoefficients;
}



/************************************************************************
THE FOLLOWING CODE IS EXTRACTED FROM SOFT10
************************************************************************/

# ifndef MAX
# define MAX(A, B) ((A) > (B) ? (A) : (B))
# endif
# define HPP(M, K, BW) ((BW) - MAX((M), (K)))
# define HPM(M, K, BW) ((BW) - MAX((M), (-K)))
# define HMP(M, K, BW) ((BW) - MAX((-M), (K)))
# define HMM(M, K, BW) ((BW) - MAX((-M), (-K)))



inline int R3WignerD::
Index(int L, int M1, int M2) const
{
  // Return index of real coefficient
  // Imaginary coefficient is at index + ncoefficients/2

  int k ;
  int tmpA, tmpB ;
  int bw = bandwidth;
  int l = L;
  int m = M1;
  int mp = M2;

  if ( m >= 0 )
    {
      if ( mp >= 0 )
	{
	  tmpA = ( 6*bw*bw + 3*m - 2*m*m - 1);
	  tmpA *= m ;
	  tmpA /= 6 ;
	  
	  tmpB = 0 ;
	  for( k = 0 ; k < mp ; k ++ )
	    tmpB += HPP(m, k, bw) ;
	  
	  tmpA += tmpB ;
	  tmpA += (l - MAX(m,mp));
	}
      else /* mp < 0 */
	{
	  tmpA = -(m*(1+m)*(1+2*m))/6 ;
	  tmpA += bw*bw*(1+m);

	  tmpB = 0;
	  for( k = mp; k < 0 ; k ++ )
	    tmpB += HPM(m, k, bw) ;

	  tmpA -= tmpB ;
	  tmpA += (l - MAX(m,-mp));
	}
    }
  else /* so m < 0 */
    {
      if ( mp >= 0 )
	{
	  tmpA = (4*bw*bw*bw-bw)/3 ;
	  tmpB = (-m)*(1-m)*(1-2*m) ;
	  tmpB /= -6 ;
	  tmpB += bw*bw*(-m) ;
	  tmpA -= tmpB ;
	  
	  tmpB = 0 ;
	  for( k = 0 ; k < mp ; k ++ )
	    tmpB += HMP(m, k, bw) ;
	  
	  tmpA += tmpB ;
	  tmpA += (l - MAX(-m,mp));
	}
      else /* mp < 0 */
	{
	  tmpA = (4*bw*bw*bw-bw)/3 ;
	  tmpB = ((-m)-1)*(6*bw*bw-m-2*(m*m));
	  tmpB /= 6 ;
	  tmpA -= tmpB ;

	  tmpB = 0;
	  for( k = mp; k < 0 ; k ++ )
	    tmpB += HMM(m, k, bw) ;

	  tmpA -= tmpB ;
	  tmpA += (l - MAX(-m,-mp));
	}
    }

  return tmpA ;
}



/******************************************************
  so3CombineCoef: combine the S^2 spherical coefficients
                  of the PATTERN h and the SIGNAL f into a
		  SO(3) coefficient array in such a way
		  so that the inverse SO(3) transform will
		  result in the correlation of f and h, and
		  the location of the max value of the
		  inverse transform will tell you the Euler
		  angles required to rotate the PATTERN h to
		  align it with the SIGNAL f

  degLim: you may want to consider coefficients through
          degLim. In which case, the coefficients for
	  degrees degLim + 1 through bwOut-1 will all be 
	  set to 0 in the SO(3) coefficient array
******************************************************/

void R3WignerD::
AddSphericalHarmonics(const R2SphericalHarmonics& harmonics1, const R2SphericalHarmonics& harmonics2)
{
  // Check resolutions
  int bw = Bandwidth();
  assert(bw == harmonics1.Bandwidth());
  assert(bw == harmonics2.Bandwidth());

  // Add spherical harmonics correlation
  int degLim = bw -1;
  for (int L = 0; L <= degLim; L++ ) {
    for (int M1 = -L; M1 <= L; M1 ++) {
      // Grab signal coefficient, at this degree and order 
      const RNScalar *h1 = harmonics1.Coefficients(L, -M1);
      RNScalar tmpSigCoefR = *h1;
      RNScalar tmpSigCoefI = *(h1 + 1);

      // Need to reset the -1 fudge factor 
      int fudge;
      if ( (M1 + L) % 2 ) fudge = -1 ;
      else fudge = 1 ;

      for (int M2 = -L ; M2 <= L ; M2 ++ ) {
        // First take the CONJUGATE of the pattern coef and multiply it by the fudge factor
        const RNScalar *h2 = harmonics2.Coefficients(L, -M2);
        RNScalar tmpPatCoefR = fudge *  (*h2);
        RNScalar tmpPatCoefI = -fudge * (*(h2 + 1));

        // Now multiply the signal coef by the pattern coef and save it in the so3 coefficient array
        int index = Index(L, M1, M2) ;
        coefficients[ index ] +=
          tmpSigCoefR*tmpPatCoefR -
          tmpSigCoefI*tmpPatCoefI ;

        coefficients [ ncoefficients/2 + index ] +=
          tmpSigCoefR*tmpPatCoefI +
          tmpSigCoefI*tmpPatCoefR ;

        // multiply fudge factor by -1
        fudge *= -1 ;
      }
    }
  }
}



int R3WignerD::
InverseWignerDTransform(double *euler_grid) const
{
  // Check resolutions
  int bw = Bandwidth();
  int size = 2 * bw;

  // Allocate temporary memory
  double *workspace1 = new double [4 * size * size * size];
  double *workspace2 = new double [24 * bw + 2 * bw * bw];
  double *so3SigI = new double [size * size * size]; 
  assert(workspace1 && workspace2 && so3SigI);

  // Compute inverse WignerD transform
  Inverse_SO3_Naive(bw, coefficients, &coefficients[ncoefficients/2], euler_grid, so3SigI, workspace1, workspace2);

  // Delete temporary memory
  delete [] workspace1;
  delete [] workspace2;
  delete [] so3SigI;

  // Return success
  return size;
}



